#include "scripts.h"

#include "logging.h"

ScriptTrader::ScriptTrader(Director &d) : Dispatch(d) {
  BUGZ_LOG(fatal) << "ScriptTrader()";
  state = 0;
};

ScriptTrader::~ScriptTrader() { BUGZ_LOG(fatal) << "~ScriptTrader()"; }

void ScriptTrader::activate(void) {
  // ok, lookup port1 port2
  BUGZ_LOG(fatal) << "ScriptTrader::activate " << port[0] << " & " << port[1];
  auto port_info = director.galaxy.ports.find(port[0]);
  int port0_type = port_info->second.type;
  port_buysell[0] = get_buysell(port0_type);
  port_info = director.galaxy.ports.find(port[1]);
  int port1_type = port_info->second.type;
  port_buysell[1] = get_buysell(port1_type);
  BUGZ_LOG(fatal) << port0_type << " and " << port1_type;
  auto ttr = trade_type_info(port0_type, port1_type);
  trades = ttr.trades;

  if (trades.foe[0] && trades.foe[1] && trades.foe[2]) {
    // it has all three -- use the last 2.
    trades.foe[0] = false;
  }

  // Ok, what do we do first here?
  // I - Info
  state = 1;
  percent = 5.0;
  to_server("I");
  director.galaxy.meta["help"]["stop_percent"] =
      "ScriptTrader stop trading if below this percent.";

  if (director.galaxy.config["stop_percent"]) {
    stop_percent = director.galaxy.config["stop_percent"].as<int>();
  } else {
    stop_percent = 20;
    director.galaxy.config["stop_percent"] = stop_percent;
  }
  director.galaxy.meta["help"]["trade_end_empty"] =
      "ScriptTrader end trades with empty holds? Y/N";

  if (director.galaxy.config["trade_end_empty"]) {
    std::string tee =
        director.galaxy.config["trade_end_empty"].as<std::string>();
    if ((toupper(tee[0]) == 'Y') || (toupper(tee[0]) == 'T')) {
      trade_end_empty = true;
    } else {
      trade_end_empty = false;
      // director.galaxy.config["trade_end_empty"] = "N";
    }
  } else {
      trade_end_empty = false;
      director.galaxy.config["trade_end_empty"] = "N";
  }
}

void ScriptTrader::deactivate(void) { notify(); }

void ScriptTrader::server_line(const std::string &line,
                               const std::string &raw_line) {
  // FUTURE:  powering up weapons check

  // Show what's going on...
  if (state > 1) {
    std::string temp = raw_line;
    temp.append("\n\r");
    to_client(temp);
  }

  if (line == "Docking...") {
    last_offer = 0;
    final_offer = 0;
    initial_offer = 0;
  }

  static std::set<std::string> success_lines = {
      "If only more honest traders would port here, we'll take them though.",
      "You will put me out of business, I'll take your offer.",
      "FINE, we'll take them, just leave!",
      "Agreed, and a pleasure doing business with you!",
      "You are a rogue! We'll take them anyway.",
      "You insult my intelligence, but we'll buy them anyway.",
      "Very well, we'll take that offer.",
      "You drive a hard bargain, but we'll take them.",
      "Done, we'll take the lot.",
      "I hate haggling, they're all yours.",
      "You are robbing me, but we'll buy them anyway.",
      "SOLD!  Come back anytime!",
      "Cheapskate.  Here, take them and leave me alone.",
      "Very well, we'll buy them.",
      "You are a shrewd trader, they're all yours.",
      "I could have twice that much in the Androcan Empire, but they're yours.",
      "Oh well, maybe I can sell these to some other fool, we'll take them.",
      "I PAID more than that!  But we'll sell them to you anyway.",
      "(Sigh) Very well, pay up and take them away.",
      "Agreed! We'll purchase them!"};

  if (success_lines.find(line) != success_lines.end()) {
    BUGZ_LOG(fatal) << "Success " << buying << " " << initial_offer << " : "
                    << last_offer;
    // calculate % ?
    BUGZ_LOG(fatal) << "% " << (float)initial_offer / (float)last_offer * 100.0;
    BUGZ_LOG(fatal) << "meta trade setting: " << percent << " for "
                    << active_port << " " << product;
    director.galaxy.meta["trade"][active_port][product] = percent;
  }

  // <P-Probe estimates your offer was

  if (startswith(line, "Agreed, ")) {
    last_offer = 0;
    final_offer = 0;
    if (director.galaxy.meta["trade"][active_port][product]) {
      percent = director.galaxy.meta["trade"][active_port][product].as<float>();
      percent += 1.0;
      BUGZ_LOG(fatal) << "Percent for " << active_port << " now " << percent;
    } else {
      BUGZ_LOG(fatal) << "using default for " << active_port;
      percent = 5.0;  // check meta for past trades information
    }
  }

  if (startswith(line, "We'll buy them for ")) {
    // I need the initial offer!
    std::string offer = line.substr(19);
    replace(offer, ",", "");
    initial_offer = stoi(offer);
    BUGZ_LOG(fatal) << "Buying, initial: " << initial_offer;
    buying = true;  // Port is buying, we are selling.
  }

  if (startswith(line, "We'll sell them for ")) {
    // I need the initial offer!
    std::string offer = line.substr(20);
    replace(offer, ",", "");
    initial_offer = stoi(offer);
    BUGZ_LOG(fatal) << "Selling, initial: " << initial_offer;
    buying = false;  // Port is selling, we are buying.
  }

  // SL: [Our final offer is 1,263 credits.]
  if (startswith(line, "Our final offer is ")) {
    // Well snap!
    std::string offer = line.substr(19);
    replace(offer, ",", "");
    final_offer = stoi(offer);
    BUGZ_LOG(fatal) << "Final offer: " << final_offer;
  }

  // SL: [You have 16,767 credits and 0 empty cargo holds.]
  // trade accepted.  if not 0 empty cargo holds -- we failed!
  // SL: [<P-Probe estimates your offer was  91.83% of best price>]
  // SL: [You have 4,046 credits and 0 empty cargo holds.]

  // this shows up at the initial docking of the port.

  if (startswith(line, "You have ")) {
    if (initial_offer != 0) {
      // Ok, the offer was possibly accepted.
      int success;
      if (buying)
        success = 0;
      else
        success = director.galaxy.meta["ship"]["holds"]["total"].as<int>();

      std::string text = std::to_string(success);
      text.append(" empty cargo holds.");
      if (endswith(line, text)) {
        BUGZ_LOG(fatal) << "Trade SUCCESS!";
        // record this action somewhere in meta.
      }
    }
  }
}

void ScriptTrader::server_prompt(const std::string &prompt) {
  // FUTURE:  check for "Surrender/Attack"

  if (at_command_prompt(prompt)) {
    if (state == 1) {
      // Ok, decision time!
      if (director.galaxy.meta["ship"]["holds"]["c"]) {
        // holds contain colonists
        to_client("ScriptTrader FAIL: holds contain colonists.");
        deactivate();
        return;
      }

      // Which port to trade with first?  examine trades
      BUGZ_LOG(fatal) << "trades: " << trades;
      BUGZ_LOG(fatal) << "port0:" << text_from_buysell(port_buysell[0]);
      BUGZ_LOG(fatal) << "port1:" << text_from_buysell(port_buysell[1]);

      // Do we have what they are buying?
      bool have_buy = false;
      int active_buy = 0;
      int active_sell = 0;
      bool all_holds_empty = false;

      // check the ship and holds here.  (MAYBE)
      int holds = director.galaxy.meta["ship"]["holds"]["total"].as<int>();
      if (director.galaxy.meta["ship"]["holds"]["empty"]) {
        if (holds == director.galaxy.meta["ship"]["holds"]["empty"].as<int>())
          all_holds_empty = true;
      }

      for (int x = 0; x < 3; ++x) {
        if (trades.foe[x]) {
          // this is what they will be trading...

          if (director.galaxy.meta["ship"]["holds"][foe[x]]) {
            // key exists ...
            have_buy = true;
            // which port is buying?
            if (port_buysell[0].foe[x]) {
              active_buy = 0;
              have_buy = true;
            } else if (port_buysell[1].foe[x]) {
              active_buy = 1;
              have_buy = true;
            }
          }

          if (!port_buysell[0].foe[x]) {
            active_sell = 0;
          } else {
            active_sell = 1;
          }
        }
      }

      if (have_buy) {
        BUGZ_LOG(fatal) << "have_buy: port " << active_buy;
        active_port = port[active_buy];
      } else {
        // which port is selling?

        // if they aren't buying what I have in my holds, should I check to see
        // if my holds are full?  This could be the "not buying" what I'm
        // setting bug!

        BUGZ_LOG(fatal) << "!have_buy: port " << active_sell;
        active_port = port[active_sell];
        // yes, this is the bug alright.
        // Ok, this shows up all the time.  I need to look at my holds!
        if (!all_holds_empty) {
          to_client(
              "I don't see any ports that are buying what we have in our "
              "holds!\n\r");
          deactivate();
          return;
        }
      }

      state = 2;
      if (director.current_sector == active_port) {
        // begin state 3
        state = 3;
        to_client("Trading...\n\r");
        to_server("PT");
        return;
      } else {
        // initiate move
        std::string move = std::to_string(active_port);
        to_client("Moving...\n\r");
        move.append("\r");
        to_server(move);
        return;
      }

      // for now...
      deactivate();
    }

    if (state == 2) {
      if (director.current_sector == active_port) {
        // We're here
        state = 3;
        to_client("Trading...\n\r");
        to_server("PT");
        return;
      } else {
        // we failed to move to where we wanted to go?!
        BUGZ_LOG(fatal) << "Expecting: " << active_port << " but got "
                        << director.current_sector;
        deactivate();
        return;
      }
    }
  }

  if (state == 3) {
    if (startswith(prompt, "How many holds of ")) {
      char selling = tolower(prompt[18]);
      for (int x = 0; x < 3; ++x) {
        if (foe[x] == selling) product = x;
      }

      if (in(prompt, " to sell ")) {
        // always sell everything
        to_server("\r");
        return;
      }

      if (in(prompt, " to buy ")) {
        bool buy_ok = true;

        if (trade_end_empty) {
          // Ok, we want to end with empty holds...
          int other_port;
          if (active_port == port[0])
            other_port = port[1];
          else
            other_port = port[0];

          // Is target port burnt?
          auto pos = director.galaxy.ports.find(other_port);
          bool burnt = false;

          if (pos != director.galaxy.ports.end()) {
            // We'll find the port.  Really.

            for (int x = 0; x < 3; ++x) {
              if (trades.foe[x]) {
                if (pos->second.percent[x] < stop_percent) burnt = true;
              }
            }
          }

          if (burnt) {
            buy_ok = false;
          }
        }

        // Ok, what are they selling?
        // char selling = tolower(prompt[18]);
        BUGZ_LOG(fatal) << "Selling: " << selling;
        if (!buy_ok) {
          // no!
          to_server("0\r");
        } else
          for (int x = 0; x < 3; ++x) {
            if (foe[x] == selling) {
              // We found the item ... is it something that we're trading?
              if (trades.foe[x]) {
                // Yes!
                to_server("\r");
                product = x;
              } else {
                // No!
                to_server("0\r");
              }
            }
          }
      }
    }

    if (startswith(prompt, "Your offer [") && endswith(prompt, " ? ")) {
      // Ok, things get weird here.  We also need to look for final offer.
      if (last_offer != 0) percent -= 1.0;

      if (buying)
        last_offer = (int)(initial_offer * (100 + percent) / 100.0);
      else
        last_offer = (int)(initial_offer * (100 - percent) / 100.0);

      BUGZ_LOG(fatal) << "Offer: " << buying << " offer " << last_offer << " % "
                      << percent;
      std::string text = std::to_string(last_offer);
      text.append("\r");
      to_server(text);
    }

    if (at_command_prompt(prompt)) {
      // we're done trading...
      // do we carry on, or stop?
      // 1.) CHECK TURNS // need turn tracking
      // 2.) PORTS BURNT?

      if (active_port == port[0])
        active_port = port[1];
      else
        active_port = port[0];

      // Is target port burnt?
      auto pos = director.galaxy.ports.find(active_port);
      bool burnt = false;

      if (pos != director.galaxy.ports.end()) {
        // We'll find the port.  Really.

        for (int x = 0; x < 3; ++x) {
          if (trades.foe[x]) {
            BUGZ_LOG(fatal) << x << " % " << (int)pos->second.percent[x] << " "
                            << stop_percent;
            if (pos->second.percent[x] < stop_percent) burnt = true;
          }
        }
      }

      if (burnt) {
        to_client("Ports burnt.\n\r");
        deactivate();
        return;
      }

      std::string move = std::to_string(active_port);
      to_client("Moving...\n\r");
      move.append("\r");
      to_server(move);

      state = 2;
    }
  }
}
void ScriptTrader::client_input(const std::string &cinput) { deactivate(); };
